"""victoria_email

A Victoria plugin for managing the FileTrust Rebuild for Email platform.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>
"""
import logging
from typing import List

import aiorun
import click
from victoria.plugin import Plugin

from victoria_email.core.util import generate_random_email

from . import load_test, schemas, reconstruct_mail, replay_deadletters, recover_mail, send_mail


def ensure_mailtoil(cfg: schemas.EmailConfig) -> None:
    """Logs an error and exits if mailtoil is not present in config."""
    if cfg.mail_toil is None:
        logging.error(
            "You need to configure the 'mail_toil' section of your config to use this command!"
        )
        raise SystemExit(1)


def ensure_loadtest(cfg: schemas.EmailConfig) -> None:
    """Logs an error and exits if loadtest is not present in config."""
    if cfg.load_test is None:
        logging.error(
            "You need to configure the 'mail_toil' section of your config to use this command!"
        )
        raise SystemExit(1)


@click.group()
def root_cmd() -> None:
    """Perform various actions on the Rebuild for Email platform."""


@root_cmd.command()
@click.option(
    "-n",
    "--frequency",
    type=int,
    required=False,
    default=1,
    help="The number of emails to send per second of the test. Default: 1.")
@click.option("-e",
              "--endpoint",
              type=str,
              required=True,
              help="The SMTP endpoint (and optional port) to send to.")
@click.option("-t",
              "--duration",
              type=int,
              required=False,
              default=1,
              help="The duration in seconds of the test. Default: 1.")
@click.option("-r",
              "--recipient",
              type=str,
              default=generate_random_email(),
              help="The email recipient address. Default: A random email recipient will be generated")
@click.option("-s",
              "--sender",
              type=str,
              required=True,
              default=generate_random_email(),
              help="The email sender address. Default: A random email sender will be generated")
@click.option("-i",
              "--tenant_ids",
              type=str,
              multiple=True,
              required=False,
              default=None,
              help="Tenant id. Default: A random tenant id will be generated")
@click.pass_obj
def loadtest(cfg: schemas.EmailConfig, frequency: int, endpoint: str,
             duration: int, recipient: str, sender: str, tenant_ids: list) -> None:
    """Perform a load test on a cluster.
    
    \b
    Send a single email:
    $ victoria email loadtest -e smtp.example.com -s test@example.com -r test@example.com

    \b
    Send 46 mails per second for 60 seconds:
    $ victoria email loadtest -e smtp.example.com -n 46 -t 60 -s test@example.com -r test@example.com

    \b
    Send using a different port than port 25:
    $ victoria email loadtest -e smtp.example.com:465 -s test@example.com -r test@example.com
    """
    loop = aiorun.get_event_loop()
    loop.set_exception_handler(lambda loop, context: "Error")
    loop.run_until_complete(
        load_test.perform_load_test(frequency, endpoint, duration, recipient,
                                    sender, tenant_ids, cfg.load_test))


@root_cmd.command()
@click.argument("cluster", nargs=1, type=str)
@click.option("-o",
              "--output",
              type=str,
              required=True,
              help="The directory to write reconstructed mail to.")
@click.option("-a", "--anon", type=bool, is_flag=True)
@click.option("-i",
              "--transaction-id",
              multiple=True,
              type=str,
              help="Cherry-pick GUIDs from blob storage to reconstruct. "
              "This bypasses dead letter scanning.")
@click.pass_obj
def reconstruct(cfg: schemas.EmailConfig, cluster: str, output: str,
                anon: bool, transaction_id: List[str]) -> None:
    """Reconstruct mail from blob storage.

    Can also scan dead letter queues to find mail that needs to be reconstructed.

    \b
    Reconstruct mail in dead letters:
    $ victoria email reconstruct uksprod1 -o output_dir

    \b
    Reconstruct a specific transaction ID
    $ victoria email reconstruct useprod2 -i <guid> -o output_dir

    \b
    Reconstruct a specific transaction ID, anonymising contents
    $ victoria email reconstruct useprod4 -i <guid> -o output_dir --anon
    """
    ensure_mailtoil(cfg)
    reconstruct_mail.reconstruct(cfg.mail_toil, cluster, output,
                                 transaction_id, anon, cfg)


@root_cmd.command()
@click.argument("cluster", nargs=1, type=str)
@click.pass_obj
def replay(cfg: schemas.EmailConfig, cluster: str) -> None:
    """Replay mail in service bus dead letters.

    \b
    To replay mail from a cluster:
    $ victoria email replay uksprod1
    """
    ensure_mailtoil(cfg)
    replay_deadletters.replay(cfg.mail_toil, cluster, cfg)


@root_cmd.command()
@click.argument("cluster", nargs=1, type=str)
@click.option("-i",
              "--input",
              metavar="FILE",
              type=str,
              required=True,
              help="TXT file containing transaction IDs to replay.")
@click.option("-o",
              "--output",
              metavar="URL",
              type=str,
              required=True,
              help="The SMTP endpoint to replay mail to.")
@click.pass_obj
def recover(cfg: schemas.EmailConfig, cluster: str, input: str,
            output: str) -> None:
    """Replay mail from blob storage through SaaS.

    To be used in the event of a failure causing mail to be persisted to
    blob storage but not sent onwards from SMTP receiver - being lost into
    the void.

    \b
    Usage example:
    $ victoria email recover uksprod -i tx-ids.txt -o localhost:25
    """
    ensure_mailtoil(cfg)
    recover_mail.recover(cfg.mail_toil, cluster, input, output, cfg)


@root_cmd.command()
@click.argument("manifest", nargs=1, type=str)
def send(manifest: str) -> None:
    """Send mail specified by manifest files.

    For information about the manifest format, please see the README:
    https://github.com/glasswall-sre/victoria_email

    \b
    Usage example:
    $ victoria email send uksprod.yaml
    """
    loaded_manifest = send_mail.Manifest.load(manifest)
    send_mail.send_manifest(loaded_manifest)


# plugin entry point
plugin = Plugin(name="email",
                cli=root_cmd,
                config_schema=schemas.EmailConfigSchema())



import string
import uuid
import random
from os.path import splitext
from typing import List, Iterable
from uuid import UUID

DOMAINS = ["hotmail.com", "gmail.com", "aol.com", "mail.com", "yahoo.com"]


def file_ext(f: str) -> str:
    """Get the file extension of a file."""
    return splitext(f)[1]


def file_name(f: str) -> str:
    """Get the filename of a file without the extension."""
    return splitext(f)[0]


def get_random_items(items: Iterable, count: int) -> list:
    iterator = iter(items)
    try:
        items = [iterator.next() for _ in range(count)]
    except StopIteration:
        raise ValueError("Sample larger than population.")
    random.shuffle(items)
    for i, v in enumerate(iterator, count):
        r = random.randint(0, i)
        if r < count:
            items[r] = v
    return items


def get_random_domain(domains: List[str]) -> str:
    return random.choice(domains)


def get_random_name(length: int) -> str:
    letters = string.ascii_lowercase[:12]
    return ''.join(random.choice(letters) for _ in range(length))


def generate_random_email(length: int = 10) -> str:
    return f'{get_random_name(length)}@{get_random_domain(DOMAINS)}'


def generate_random_uuids(count: int = 10) -> List[UUID]:
    return [uuid.uuid4() for _ in range(count)]


def generate_random_numbers():
    return [random.randint(1, 100) for _ in range(0, 5)]



"""recover_mail

The functionality of the recover MailToil script is ported here.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>
"""
import email
import logging
import logging.config
import os
from os.path import join
from typing import List
import smtplib

from .core import config, service_bus, blob_storage, mail_reconstruction
from .schemas import EmailConfig


def recover(cfg: config.MailToilConfig, cluster: str, input_file: str,
            smtp_addr: str, plugin_cfg: EmailConfig) -> None:
    """Perform the mailtoil recover functionality.

    Args:
        cfg: The mail toil config.
        cluster: The cluster to recover.
        input_file: The file path to the input tx ID txt file.
        smtp_addr: The SMTP address to send to.
        plugin_cfg: The email plugin config object.
    """
    encryption_provider = plugin_cfg.victoria_config.get_encryption()
    storage_conn_str = encryption_provider.decrypt_str(
        cfg.get_storage_account(cluster))
    if storage_conn_str is None:
        raise SystemExit(1)

    # connect to blob
    blob_client = blob_storage.connect(cfg.get_storage_account(cluster))

    tx_ids = []
    with open(input_file, "r") as tx_id_file:
        for line in tx_id_file:
            tx_ids.append(line.strip())

    with smtplib.SMTP(smtp_addr) as smtp:
        for tx_id in tx_ids:
            mime_str = blob_storage.get_mime_message(tx_id, blob_client)
            print(f"Found MIME message for tx '{tx_id}'")
            preprocessed_msg = mail_reconstruction.preprocess_mime_message(
                mime_str).encode("utf-8")
            mime_msg = email.message_from_bytes(preprocessed_msg)
            smtp.send_message(mime_msg)
            print(f"Sent to: {mime_msg['To']}, from: {mime_msg['From']}")


"""schemas

Marshmallow schemas for the email plugin.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>
"""
from dataclasses import dataclass, field
from typing import Optional
from uuid import UUID
from typing import List, Dict

from marshmallow import Schema, fields, post_load, validate

from .core.blob_storage import CONNECTION_STR, get_blob_properties
from .core.util import get_random_items
from .core.config import MailToilConfigSchema, MailToilConfig


class Distribution:
    def __init__(self, file: str, weight: float):
        self.file = file
        self.weight = weight

    @classmethod
    def get_random_distributions(cls):
        properties = get_blob_properties('fileattachments', CONNECTION_STR)
        return [cls(attachment.name, attachment.size) for attachment in
                get_random_items(properties, count=100)]


class Load:
    def __init__(self, distribution: List[Dict] = None, attachment_count: List[int] = None):
        self.distribution = distribution
        self.attachment_count = attachment_count


class Function:
    def __init__(self, function: str, mail_send_function_code: str):
        self.function = function
        self.mail_send_function_code = mail_send_function_code


class DistributionSchema(Schema):
    file = fields.Str(required=True)
    weight = fields.Float(required=True)

    @post_load
    def make_config(self, data, **kwargs):
        return Distribution(**data)


class FunctionSchema(Schema):
    function = fields.Str(required=True)
    mail_send_function_code = fields.Str(required=True)

    @post_load
    def make_config(self, data, **kwargs):
        return Function(**data)

    def __str__(self):
        return f'{self.function}'


class LoadSchema(Schema):
    distribution = fields.List(fields.Nested(DistributionSchema))
    attachment_count = fields.List(fields.Int())

    @post_load
    def make_config(self, data, **kwargs):
        return Load(**data)


class LoadTestConfigSchema(Schema):
    """Marshmallow schema for the load testing config section."""
    mail_send_function_endpoints = fields.List(fields.Nested(FunctionSchema), required=True)
    # mail_send_function_code = fields.Str(required=True, allow_none=False)
    tenant_ids = fields.List(fields.UUID(allow_none=False), required=False)
    timeout = fields.Float(required=False, allow_none=False, missing=1.0)
    load = fields.Nested(LoadSchema, required=False)

    @post_load
    def make_config(self, data, **kwargs):
        return LoadTestConfig(**data)


@dataclass
class LoadTestConfig:
    """The config of the load tester.

    Attributes:
        mail_send_function_endpoints: The HTTP endpoints of the going-postal backend.
        tenant_ids: The tenant ID(s) to attach to the sent tests.
        timeout: The SMTP sending timeout to use.
        load:
    """
    # mail_send_function_code: str
    timeout: float
    mail_send_function_endpoints: field(default_factory=Function)
    load: Load = field(default_factory=Load)
    tenant_ids: List[UUID] = field(default_factory=list)


class EmailConfigSchema(Schema):
    """Marshmallow schema for the email plugin config."""
    load_test = fields.Nested(LoadTestConfigSchema,
                              required=False,
                              allow_none=True,
                              missing=None)
    mail_toil = fields.Nested(MailToilConfigSchema,
                              required=False,
                              allow_none=True,
                              missing=None)

    @post_load
    def make_config(self, data, **kwargs):
        return EmailConfig(**data)


@dataclass
class EmailConfig:
    """The email plugin config.

    Attributes:
        load_test: The config for the load testing command.
        mail_toil: The config for the mail toil commands.
    """
    load_test: LoadTestConfig
    mail_toil: Optional[MailToilConfig]



"""reconstruct_mail

Contains the functionality for the mail reconstruction portion of the
mailtoil scripts.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>
"""
import datetime
import email
import logging
from os import makedirs
from typing import List, Optional

from .core import blob_storage
from .core import config
from .core import mail_reconstruction
from .core import service_bus
from .schemas import EmailConfig


def create_output_dir(output_dir: str) -> None:
    """Create the output directory."""
    try:
        makedirs(output_dir)
    except FileExistsError:
        # don't do anything if it already existed
        pass


def get_dead_letters_from_service_bus(cluster: str, cfg: config.MailToilConfig,
                                      conn_str: str) -> List[str]:
    """Get the list of dead letter IDs from the service bus."""
    # connect to the service bus for the cluster

    sb_client = service_bus.connect(
        cfg.get_service_bus_connection_str(cluster))

    # scan for dead letters to get IDs to cross reference in blob storage
    transaction_ids = []
    logging.info(f"Scanning queues on '{cluster}' for dead letters...")
    for queue in cfg.queues:
        dead_letters = service_bus.get_all_dead_letter_ids(queue, sb_client)
        for dead_letter in dead_letters:
            logging.info(f"\tFound dead letter '{dead_letter}'")
        transaction_ids += dead_letters
    logging.info(f"Found {len(transaction_ids)} dead letter(s) on '{cluster}'")
    return transaction_ids


def reconstruct(cfg: config.MailToilConfig, cluster: str, output_dir: str,
                transaction_ids: List[str], anonymise: bool,
                plugin_cfg: EmailConfig) -> None:
    """Perform the reconstruct functionality.

    Args:
        cfg: The mail toil config.
        cluster: The cluster to grab transactions from.
        output_dir: Where to write the reconstructed mail to.
        transaction_ids: The tx IDs to reconstruct.
        anonymise: Whether we should anonymise.
        plugin_cfg: The email plugin config object.
    """
    encryption_provider = plugin_cfg.victoria_config.get_encryption()
    service_bus_conn_str = encryption_provider.decrypt_str(
        cfg.get_service_bus_connection_str(cluster))
    storage_conn_str = encryption_provider.decrypt_str(
        cfg.get_storage_account(cluster))
    if service_bus_conn_str is None or storage_conn_str is None:
        raise SystemExit(1)

    create_output_dir(output_dir)

    # if transaction IDs weren't given then get them from dead letters
    # instead
    if len(transaction_ids) == 0:
        transaction_ids = get_dead_letters_from_service_bus(
            cluster, cfg, service_bus_conn_str)

    # now grab the MIME messages of these transactions and reconstruct them
    logging.info(f"Connecting to blob storage for '{cluster}'...")
    blob_client = blob_storage.connect(storage_conn_str)

    for transaction_id in transaction_ids:
        mime_msg = blob_storage.get_mime_message(transaction_id, blob_client)
        if mime_msg is not None:
            mail_reconstruction.process_mime_message(mime_msg, transaction_id,
                                                     output_dir, anonymise)



"""victoria_email.core

This is the core functionality of the email plugin.

Code in this package is converted one-to-one from numerous disparate scripts,
so may not fit the formatting/style of other code.

You may have previously known this functionality as 'MailToil'.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>
"""